@@ -36,12 +36,16 @@ http://wiki.secondlife.com/wiki/LLSD
 
 import base64
 import binascii
+import calendar
 import datetime
 import re
 import struct
 import time
 import types
 import uuid
+import urlparse
+import os
+import exceptions
 
 from fastest_elementtree import ElementTreeError, fromstring
 
@@ -74,6 +78,8 @@ class uri(str):
 _int_regex = re.compile(r"[-+]?\d+")
 _real_regex = re.compile(r"[-+]?(?:(\d+(\.\d*)?|\d*\.\d+)([eE][-+]?\d+)?)|[-+]?inf|[-+]?nan")
 _alpha_regex = re.compile(r"[a-zA-Z]+")
+_true_regex = re.compile(r"TRUE|true|\b[Tt]\b")
+_false_regex = re.compile(r"FALSE|false|\b[Ff]\b")
 _date_regex = re.compile(r"(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})T"
                         r"(?P<hour>\d{2}):(?P<minute>\d{2}):(?P<second>\d{2})"
                         r"(?P<second_float>(\.\d+)?)Z")
@@ -84,10 +90,13 @@ def _format_datestr(v):
     Formats a datetime or date object into the string format shared by
     xml and notation serializations.
     """
-    if hasattr(v, 'microsecond'):
-        return v.isoformat() + 'Z'
-    else:
-        return v.strftime('%Y-%m-%dT%H:%M:%SZ')
+    if not isinstance(v, datetime.date) and not isinstance(v, datetime.datetime):
+        raise LLSDParseError("invalid date string %s passed to date formatter" % s)
+    
+    if not isinstance(v, datetime.datetime):
+        v = datetime.datetime.combine(v, datetime.time(0))
+    
+    return v.isoformat() + 'Z'
 
 def _parse_datestr(datestr):
     """
@@ -147,11 +156,22 @@ def _str_to_python(node):
     return node.text or ''
 
 def _bin_to_python(node):
+    base = node.get('encoding') or 'base64'
     try:
-        return binary(base64.decodestring(node.text or ''))
+        if base == 'base16':
+            # parse base16 encoded data
+            return binary(base64.b16decode(node.text or ''))
+        elif base == 'base64':
+            # parse base64 encoded data
+            return binary(base64.b64decode(node.text or ''))
+        elif base == 'base85':
+            return LLSDParseError("Parser doesn't support base85 encoding")
     except binascii.Error, exc:
         # convert exception class so it's more catchable
-        raise LLSDParseError("Base64 " + str(exc))
+        return LLSDParseError("Encoded binary data: " + str(exc))
+    except exceptions.TypeError, exc:
+        # convert exception class so it's more catchable
+        return LLSDParseError("Bad binary data: " + str(exc))
 
 def _date_to_python(node):
     "Convert date node to a python object."
@@ -291,7 +311,7 @@ class LLSDXMLFormatter(object):
         else:
             return self._elt('uuid', v)
     def BINARY(self, v):
-        return self._elt('binary', base64.encodestring(v))
+        return self._elt('binary', base64.b64encode(v).strip())
     def STRING(self, v):
         return self._elt('string', self.xml_esc(v))
     def URI(self, v):
@@ -496,7 +516,8 @@ class LLSDNotationFormatter(object):
     def UUID(self, v):
         return "u%s" % v
     def BINARY(self, v):
-        return 'b64"' + base64.encodestring(v) + '"'
+        return 'b64"' + base64.b64encode(v).strip() + '"'
+
     def STRING(self, v):
         if isinstance(v, unicode):
             v = v.encode('utf-8')
@@ -516,7 +537,7 @@ class LLSDNotationFormatter(object):
              for key, value in v.items()])
 
     def _generate(self, something):
-        "Generate xml from a single python object."
+        "Generate notation from a single python object."
         t = type(something)
         handler = self.type_map.get(t)
         if handler:
@@ -630,11 +651,11 @@ class LLSDBinaryParser(object):
             idx = self._index
             self._index += 8
             seconds = struct.unpack("!d", self._buffer[idx:idx+8])[0]
-            return datetime.datetime.fromtimestamp(seconds)
+            return datetime.datetime.utcfromtimestamp(seconds)
         elif cc == 'b':
-            binary = self._parse_string_raw()
+            b = binary(self._parse_string_raw())
             if self._keep_binary:
-                return binary
+                return b
             # *NOTE: maybe have a binary placeholder which has the
             # length.
             return None
@@ -796,25 +817,70 @@ class LLSDNotationParser(object):
         self._index = 0
         return self._parse()
 
+    def _error(self, message):
+        raise LLSDParseError("%s at index %d" % (message, self._index))
+
+    def _peek(self):
+        if self._index >= len(self._buffer):
+            self._error("unexpected end of data")
+        return self._buffer[self._index]
+
+    def _getc(self, num=1):
+        if self._index == len(self._buffer):
+            self._error("Unexpected end of data")
+        if self._index + num > len(self._buffer):
+            self._error("Trying to read past end of buffer")
+        chars = self._buffer[self._index:self._index + num]
+        self._index += num
+        return chars
+
+    def _get_until(self, delim):
+        start = self._index
+        end = self._buffer.find(delim, start)
+        if end == -1:
+            return None
+        else:
+            self._index = end + 1
+            return self._buffer[start:end]
+
+    def _get_re(self, regex):
+        match = re.match(regex, self._buffer[self._index:])
+        if not match:
+            return None
+        else:
+            self._index += match.end()
+            return match.group(0)
+
     def _parse(self):
         "The notation parser workhorse."
-        cc = self._buffer[self._index]
-        self._index += 1
+        cc = self._peek()
         if cc == '{':
+            # map
             return self._parse_map()
         elif cc == '[':
-            return self._parse_array()
+            # array
+           return self._parse_array()
         elif cc == '!':
+            # undefined
+            self._getc() # eat the '!'
             return None
         elif cc == '0':
+            # false
+            self._getc() # eat the '0'
             return False
         elif cc == '1':
+            # true
+            self._getc() # eat the '1'
             return True
         elif cc in ('F', 'f'):
-            self._skip_alpha()
+            # false, must check for F|f|false|FALSE
+            if self._get_re(_false_regex) is None:
+                self._error("Expected 'false' token")
             return False
         elif cc in ('T', 't'):
-            self._skip_alpha()
+            # true, must check for T|t|true|TRUE
+            if self._get_re(_true_regex) is None:
+                self._error("Expected 'true' token")
             return True
         elif cc == 'i':
             # 'i' = integer
@@ -826,41 +892,80 @@ class LLSDNotationParser(object):
             # 'u' = uuid
             return self._parse_uuid()
         elif cc in ("'", '"', 's'):
-            return self._parse_string(cc)
+            # string
+            return self._parse_string()
         elif cc == 'l':
             # 'l' = uri
-            delim = self._buffer[self._index]
-            self._index += 1
-            val = uri(self._parse_string(delim))
-            return val
+            return self._parse_uri()
         elif cc == ('d'):
             # 'd' = date in seconds since epoch
             return self._parse_date()
         elif cc == 'b':
             return self._parse_binary()
-        else:
-            raise LLSDParseError("invalid token at index %d: %d" % (
-                self._index - 1, ord(cc)))
+       
+        # output error if the token wasn't handled
+        self._error("Invalid token %d" % ord(cc))
 
     def _parse_binary(self):
         "parse a single binary object."
-        i = self._index
-        if self._buffer[i:i+2] == '64':
-            q = self._buffer[i+2]
-            e = self._buffer.find(q, i+3)
-            if e == -1:
-                raise LLSDParseError('Unterminated binary at byte %d' % i,)
-            try:
-                try:
-                    return base64.decodestring(self._buffer[i+3:e])
-                except binascii.Error, exc:
-                    # convert exception class so it's more catchable
-                    raise LLSDParseError("Base64 " + str(exc))
-            finally:
-                self._index = e + 1
-        else:
-            raise LLSDParseError('random horrible binary format not supported')
+        
+        self._getc()    # eat the beginning 'b'
+        cc = self._peek()
+        if cc == '(':
+            # parse raw binary
+            paren = self._getc()
 
+            # grab the 'expected' size of the binary data
+            size = self._get_undil(')')
+            if size == None:
+                self._error("Invalid binary size")
+            size = int(size)
+
+            # grab the opening quote
+            q = self._getc()
+            if q != '"':
+                self._error('Expected " to start binary value')
+
+            # grab the data
+            data = self._getc(size)
+
+            # grab the closing quote
+            q = self._getc()
+            if q != '"':
+                self._error('Expected " to end binary value')
+
+            return binary(data)
+
+        else:
+            # get the encoding base
+            base = self._getc(2)
+            if base not in ('16', '64'):
+                self._error('Found unsupported binary encoding')
+
+            # grab the double quote
+            q = self._getc()
+            if q != '"':
+                self._error('Expected " to start binary value')
+
+            # grab the encoded data
+            encoded = self._get_until(q)
+
+            try:
+                if base == '16':
+                    # parse base16 encoded data
+                    return binary(base64.b16decode(encoded or ''))
+                elif base == '64':
+                    # parse base64 encoded data
+                    return binary(base64.b64decode(encoded or ''))
+                elif base == '85':
+                    self._error("Parser doesn't support base85 encoding")
+            except binascii.Error, exc:
+                # convert exception class so it's more catchable
+                self._error("Encoded binary data: " + str(exc))
+            except exceptions.TypeError, exc:
+                # convert exception class so it's more catchable
+                self._error("Bad binary data: " + str(exc))
+    
     def _parse_map(self):
         """
         parse a single map
@@ -868,33 +973,36 @@ class LLSDNotationParser(object):
         map: { string:object, string:object }
         """
         rv = {}
-        cc = self._buffer[self._index]
-        self._index += 1
         key = ''
         found_key = False
+        self._getc()   # eat the beginning '{'
+        cc = self._peek()
         while (cc != '}'):
+            if cc is None:
+                self._error("Unclosed map")
             if not found_key:
                 if cc in ("'", '"', 's'):
-                    key = self._parse_string(cc)
+                    key = self._parse_string()
                     found_key = True
                 elif cc.isspace() or cc == ',':
+                    self._getc()    # eat the character
                     pass
                 else:
-                    raise LLSDParseError("invalid map key at byte %d." % (
-                                        self._index - 1,))
+                    self._error("Invalid map key")
             elif cc.isspace():
+                self._getc()    # eat the space
                 pass
-            elif  cc == ':':
+            elif cc == ':':
+                self._getc()    # eat the ':'
                 value = self._parse()
                 rv[key] = value
                 found_key = False
             else:
-                raise LLSDParseError(
-                    "missing separator around byte %d." % (self._index - 1,))
-            cc = self._buffer[self._index]
-            self._index += 1
+                self._error("missing separator")
+            cc = self._peek()
 
-                
+        if self._getc() != '}':
+            self._error("Invalid map close token")
 
         return rv
 
@@ -905,81 +1013,72 @@ class LLSDNotationParser(object):
         array: [ object, object, object ]
         """
         rv = []
-        cc = self._buffer[self._index]
+        self._getc()    # eat the beginning '['
+        cc = self._peek()
         while (cc != ']'):
+            if cc is None:
+                self._error('Unclosed array')
             if cc.isspace() or cc == ',':
-                self._index += 1
-                cc = self._buffer[self._index]
+                self._getc()
+                cc = self._peek()
                 continue
             rv.append(self._parse())
-            cc = self._buffer[self._index]
+            cc = self._peek()
 
-        if cc != ']':
-            raise LLSDParseError("invalid array close token at index %d." % (
-                self._index,))
-        self._index += 1
+        if self._getc() != ']':
+            self._error("Invalid array close token")
         return rv
 
     def _parse_uuid(self):
         "Parse a uuid."
-        start = self._index
-        self._index += 36
-        return uuid.UUID(hex=self._buffer[start:self._index])
+        self._getc()    # eat the beginning 'u'
+        return uuid.UUID(hex=self._getc(36))
+
+    def _parse_uri(self):
+        "Parse a URI."
+        self._getc()    # eat the beginning 'l'
+        return uri(self._parse_string())
 
-    def _skip_alpha(self):
-        match = re.match(_alpha_regex, self._buffer[self._index:])
-        if match:
-            self._index += match.end()
-            
     def _parse_date(self):
         "Parse a date."
-        delim = self._buffer[self._index]
-        self._index += 1
-        datestr = self._parse_string(delim)
+        self._getc()    # eat the beginning 'd'
+        datestr = self._parse_string()
         return _parse_datestr(datestr)
 
     def _parse_real(self):
         "Parse a floating point number."
-        match = re.match(_real_regex, self._buffer[self._index:])
-        if not match:
-            raise LLSDParseError("invalid real token at index %d." % self._index)
-
-        (start, end) = match.span()
-        start += self._index
-        end += self._index
-        self._index = end
-        return float( self._buffer[start:end] )
+        self._getc()    # eat the beginning 'r'
+        match = self._get_re(_real_regex)
+        if match == None:
+            self._error("Invalid real token")
+        return float(match)
 
     def _parse_integer(self):
         "Parse an integer."
-        match = re.match(_int_regex, self._buffer[self._index:])
-        if not match:
-            raise LLSDParseError("invalid integer token at index %d." % self._index)
+        self._getc()    # eat the beginning 'i'
+        match = self._get_re(_int_regex)
+        if match == None:
+            self._error("Invalid integer token")
+        return int(match)
 
-        (start, end) = match.span()
-        start += self._index
-        end += self._index
-        self._index = end
-        return int( self._buffer[start:end] )
-
-    def _parse_string(self, delim):
+    def _parse_string(self):
         """
         Parse a string
 
         string: "g\'day" | 'have a "nice" day' | s(size)"raw data"
         """
         rv = ""
-
+        delim = self._peek()
         if delim in ("'", '"'):
-            rv = self._parse_string_delim(delim)
+            rv = self._parse_string_delim()
         elif delim == 's':
             rv = self._parse_string_raw()
         else:
-            raise LLSDParseError("invalid string token at index %d." % self._index)
+            self._error("invalid string token")
 
         return rv
 
-    def _parse_string_delim(self, delim):
+    def _parse_string_delim(self):
         """
         Parse a delimited string
 
@@ -990,9 +1089,9 @@ class LLSDNotationParser(object):
         found_hex = False
         found_digit = False
         byte = 0
+        delim = self._getc()    # get the delimeter
         while True:
-            cc = self._buffer[self._index]
-            self._index += 1
+            cc = self._getc()
             if found_escape:
                 if found_hex:
                     if found_digit:
@@ -1043,36 +1142,30 @@ class LLSDNotationParser(object):
 
         string: s(size)"raw data"
         """ 
+        self._getc()    # eat the beginning 's'
         # Read the (size) portion.
-        cc = self._buffer[self._index]
-        self._index += 1
+        cc = self._getc()
         if cc != '(':
-            raise LLSDParseError("invalid string token at index %d." % self._index)
+            self._error("Invalid string token")
 
-        rparen = self._buffer.find(')', self._index)
-        if rparen == -1:
-            raise LLSDParseError("invalid string token at index %d." % self._index)
+        size = self._get_until(')')
+        if size == None:
+            self._error("Invalid string size")
+        size = int(size)
 
-        size = int(self._buffer[self._index:rparen])
-
-        self._index = rparen + 1
-        delim = self._buffer[self._index]
-        self._index += 1
+        delim = self._getc()
         if delim not in ("'", '"'):
-            raise LLSDParseError("invalid string token at index %d." % self._index)
+            self._error("Invalid string token")
 
-        rv = self._buffer[self._index:(self._index + size)]
-        self._index += size
-        cc = self._buffer[self._index]
-        self._index += 1
+        rv = self._getc(size)
+        cc = self._getc()
         if cc != delim:
-            raise LLSDParseError("invalid string token at index %d." % self._index)
+            self._error("Invalid string closure token")
         try:
             return rv.decode('utf-8')
         except UnicodeDecodeError, exc:
             raise LLSDParseError(exc)
 
-        
 def format_binary(something):
     """
     Format application/llsd+binary to a python object.
@@ -1124,8 +1217,12 @@ def _format_binary_recurse(something):
         return 's' + struct.pack('!i', len(something)) + something
     elif isinstance(something, uri):
         return 'l' + struct.pack('!i', len(something)) + something
-    elif isinstance(something, (datetime.datetime, datetime.date)):
-        seconds_since_epoch = time.mktime(something.timetuple())
+    elif isinstance(something, datetime.datetime):
+        seconds_since_epoch = calendar.timegm(something.utctimetuple()) \
+                              + something.microsecond / 1e6
+        return 'd' + struct.pack('!d', seconds_since_epoch)
+    elif isinstance(something, datetime.date):
+        seconds_since_epoch = calendar.timegm(something.timetuple())
         return 'd' + struct.pack('!d', seconds_since_epoch)
     elif isinstance(something, (list, tuple)):
         return _format_list(something)
@@ -1147,6 +1244,14 @@ def _format_binary_recurse(something):
                 "Cannot serialize unknown type: %s (%s)" %
                 (type(something), something))
 
+def _startswith(startstr, something):
+    if hasattr(something, 'startswith'):
+        return something.startswith(startstr)
+    else:
+        pos = something.tell()
+        s = something.read(len(startstr))
+        something.seek(pos, os.SEEK_SET)
+        return (s == startstr)
 
 def parse_binary(something):
     """
@@ -1155,7 +1260,7 @@ def parse_binary(something):
     :param something: The data to parse in an indexable sequence.
     :returns: Returns a python object.
     """
-    if something.startswith('<?llsd/binary?>'):
+    if _startswith('<?llsd/binary?>', something):
         just_binary = something.split('\n', 1)[1]
     else:
         just_binary = something
@@ -1208,10 +1313,10 @@ def parse(something, mime_type = None):
     #    return parse_notation(something)
     try:
         something = something.lstrip()   #remove any pre-trailing whitespace
-        if something.startswith('<?llsd/binary?>'):
+        if _startswith('<?llsd/binary?>', something):
             return parse_binary(something)
         # This should be better.
-        elif something.startswith('<'):
+        elif _startswith('<', something):
             return parse_xml(something)
         else:
             return parse_notation(something)
@@ -1224,7 +1329,7 @@ class LLSD(object):
         self.thing = thing
 
     def __str__(self):
-        return self.toXML(self.thing)
+        return self.as_xml(self.thing)
 
     parse = staticmethod(parse)
     as_xml = staticmethod(format_xml)
